@@ -73,7 +73,7 @@ typedef struct _args_t
     int ngt_arr, nrec;
     trio_t *trios;
     int ntrios, mtrios;
-    int output_type;
+    int output_type, clevel;
     char *output_fname;
     bcf_srs_t *sr;
 }
@@ -145,27 +145,27 @@ const char *usage(void)
         "About: Count Mendelian consistent / inconsistent genotypes.\n"
         "Usage: bcftools +mendelian [Options]\n"
         "Options:\n"
-        "   -c, --count                 count the number of consistent sites [DEPRECATED, use `-m c` instead]\n"
-        "   -d, --delete                delete inconsistent genotypes (set to \"./.\") [DEPRECATED, use `-m d` instead]\n"
-        "   -l, --list [+x]             list consistent (+) or inconsistent (x) sites [DEPRECATED, use `-m +` or `-m x` instead]\n"
-        "   -m, --mode [+acdux]         output mode (the default is `-m c`):\n"
-        "                                   + .. list consistent sites\n"
-        "                                   a .. add INFO/MERR annotation with the number of inconsistent samples\n"
-        "                                   c .. print counts, a text summary with the number of errors per trio\n"
-        "                                   d .. delete inconsistent genotypes (set to \"./.\")\n"
-        "                                   u .. list uninformative sites\n"
-        "                                   x .. list inconsistent sites\n"
-        "   -o, --output <file>         write output to a file [standard output]\n"
-        "   -O, --output-type <type>    'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n"
-        "   -r, --rules <assembly>[?]   predefined rules, 'list' to print available settings, append '?' for details\n"
-        "   -R, --rules-file <file>     inheritance rules, see example below\n"
-        "   -t, --trio <m,f,c>          names of mother, father and the child\n"
-        "   -T, --trio-file <file>      list of trios, one per line (mother,father,child)\n"
-        "   -p, --ped <file>            PED file\n"
+        "   -c, --count                     Count the number of consistent sites [DEPRECATED, use `-m c` instead]\n"
+        "   -d, --delete                    Delete inconsistent genotypes (set to \"./.\") [DEPRECATED, use `-m d` instead]\n"
+        "   -l, --list [+x]                 List consistent (+) or inconsistent (x) sites [DEPRECATED, use `-m +` or `-m x` instead]\n"
+        "   -m, --mode [+acdux]             Output mode (the default is `-m c`):\n"
+        "                                       + .. list consistent sites\n"
+        "                                       a .. add INFO/MERR annotation with the number of inconsistent samples\n"
+        "                                       c .. print counts, a text summary with the number of errors per trio\n"
+        "                                       d .. delete inconsistent genotypes (set to \"./.\")\n"
+        "                                       u .. list uninformative sites\n"
+        "                                       x .. list inconsistent sites\n"
+        "   -o, --output FILE               Write output to a file [standard output]\n"
+        "   -O, --output-type u|b|v|z[0-9]  u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n"
+        "   -r, --rules ASSEMBLY[?]         Predefined rules, 'list' to print available settings, append '?' for details\n"
+        "   -R, --rules-file FILE           Inheritance rules, see example below\n"
+        "   -t, --trio M,F,C                Names of mother, father and the child\n"
+        "   -T, --trio-file FILE            List of trios, one per line (mother,father,child)\n"
+        "   -p, --ped FILE                  PED file\n"
         "\n"
         "Example:\n"
         "   # Default inheritance patterns, override with -r\n"
-        "   #   region  mothernal_ploidy + paternal > offspring\n"
+        "   #   region  maternal_ploidy + paternal > offspring\n"
         "   X:1-60000            M/M + F > M\n"
         "   X:1-60000            M/M + F > M/F\n"
         "   X:2699521-154931043  M/M + F > M\n"
@@ -243,7 +243,7 @@ static int parse_rules(const char *line, char **chr_beg, char **chr_end, uint32_
     rule_t *rule = (rule_t*) payload;
     memset(rule, 0, sizeof(rule_t));
 
-    // mothernal ploidy
+    // maternal ploidy
     se = ss;
     while ( *se && !isspace(*se) ) se++;
     int err = 0;
@@ -259,7 +259,7 @@ static int parse_rules(const char *line, char **chr_beg, char **chr_end, uint32_
         else err = 1;
     }
     else err = 1;
-    if ( err ) error("Could not parse the mothernal ploidy, only \"M\", \"M/M\" and \".\" currently supported: %s\n",line);
+    if ( err ) error("Could not parse the maternal ploidy, only \"M\", \"M/M\" and \".\" currently supported: %s\n",line);
 
     // skip "+"
     while ( *se && isspace(*se) ) se++;
@@ -321,7 +321,12 @@ void parse_ped(args_t *args, char *fname)
         int ifather = bcf_hdr_id2int(args->hdr,BCF_DT_SAMPLE,&str.s[off[2]]);
         int imother = bcf_hdr_id2int(args->hdr,BCF_DT_SAMPLE,&str.s[off[3]]);
         int ichild = bcf_hdr_id2int(args->hdr,BCF_DT_SAMPLE,&str.s[off[1]]);
-        if ( ( ifather<0 && imother<0 ) || ichild<0 ) continue;
+
+        // The code in process() makes an attempt to work with partial families,
+        // the support is not complete though and can lead to core dumps. Therefore
+        // enforcing full trios for now.
+        // if ( ( ifather<0 && imother<0 ) || ichild<0 ) continue;
+        if ( ifather<0 || imother<0 || ichild<0 ) continue;
 
         args->ntrios++;
         hts_expand0(trio_t,args->ntrios,args->mtrios,args->trios);
@@ -331,6 +336,7 @@ void parse_ped(args_t *args, char *fname)
         trios->ichild  = ichild;
 
     } while ( hts_getline(fp, KS_SEP_LINE, &str)>=0 );
+    if ( !args->ntrios ) error("No complete trios found in the PED and VCF\n");
 
     free(str.s);
     free(off);
@@ -343,6 +349,7 @@ int run(int argc, char **argv)
     memset(&args,0,sizeof(args_t));
     args.mode = 0;
     args.output_fname = "-";
+    args.clevel = -1;
 
     static struct option loptions[] =
     {
@@ -360,6 +367,7 @@ int run(int argc, char **argv)
         {0,0,0,0}
     };
     int c;
+    char *tmp;
     while ((c = getopt_long(argc, argv, "?ht:T:p:l:m:cdr:R:o:O:",loptions,NULL)) >= 0)
     {
         switch (c)
@@ -371,8 +379,17 @@ int run(int argc, char **argv)
                           case 'u': args.output_type = FT_BCF; break;
                           case 'z': args.output_type = FT_VCF_GZ; break;
                           case 'v': args.output_type = FT_VCF; break;
-                          default: error("The output type \"%s\" not recognised\n", optarg);
+                          default:
+                          {
+                              args.clevel = strtol(optarg,&tmp,10);
+                              if ( *tmp || args.clevel<0 || args.clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                          }
                       };
+                      if ( optarg[1] )
+                      {
+                          args.clevel = strtol(optarg+1,&tmp,10);
+                          if ( *tmp || args.clevel<0 || args.clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
+                      }
                       break;
             case 'R': rules_fname = optarg; break;
             case 'r': rules_string = optarg; break;
@@ -441,7 +458,9 @@ int run(int argc, char **argv)
     args.hdr = bcf_sr_get_header(args.sr, 0);
     if ( args.mode!=MODE_COUNT )
     {
-        args.out_fh = hts_open(args.output_fname,hts_bcf_wmode2(args.output_type,args.output_fname));
+        char wmode[8];
+        set_wmode(wmode,args.output_type,args.output_fname,args.clevel);
+        args.out_fh = hts_open(args.output_fname ? args.output_fname : "-", wmode);
         if ( args.out_fh == NULL ) error("Can't write to \"%s\": %s\n", args.output_fname, strerror(errno));
         if ( args.mode&MODE_ANNOTATE )
             bcf_hdr_append(args.hdr, "##INFO=<ID=MERR,Number=1,Type=Integer,Description=\"Mendelian genotype errors\">");
@@ -459,6 +478,9 @@ int run(int argc, char **argv)
         args.trios[0].imother = bcf_hdr_id2int(args.hdr, BCF_DT_SAMPLE, list[0]);
         args.trios[0].ifather = bcf_hdr_id2int(args.hdr, BCF_DT_SAMPLE, list[1]);
         args.trios[0].ichild  = bcf_hdr_id2int(args.hdr, BCF_DT_SAMPLE, list[2]);
+        if ( args.trios[0].imother<0 ) error("The sample is not present in the VCF: %s\n",list[0]);
+        if ( args.trios[0].ifather<0 ) error("The sample is not present in the VCF: %s\n",list[1]);
+        if ( args.trios[0].ichild<0 )  error("The sample is not present in the VCF: %s\n",list[2]);
         for (i=0; i<n; i++) free(list[i]);
         free(list);
     }
