@@ -74,7 +74,8 @@ typedef struct
 {
     convert_t *convert;
     filter_t *filter;
-    int argc, filter_logic, regions_is_file, targets_is_file, list_hdr, record_cmd_line;
+    int argc, filter_logic, regions_is_file, targets_is_file, list_hdr, record_cmd_line, clevel;
+    int regions_overlap, targets_overlap;
     kstring_t kstr;
     char *filter_str,
         *vep_tag;       // the --annotation INFO tag to process
@@ -115,6 +116,7 @@ typedef struct
     col2type_t *column2type;
     int ncolumn2type;
     int raw_vep_request;        // raw VEP tag requested and will need subsetting
+    int allow_undef_tags;
 }
 args_t;
 
@@ -194,15 +196,16 @@ static const char *usage_text(void)
         "       --columns-types -|FILE      Pass \"-\" to print the default -c types or FILE to override the presets\n"
         "   -d, --duplicate                 Output per transcript/allele consequences on a new line rather rather than\n"
         "                                     as comma-separated fields on a single line\n"
-        "   -f, --format <string>           Formatting expression for non-VCF/BCF output, same as `bcftools query -f`\n"
+        "   -f, --format STR                Create non-VCF output; similar to `bcftools query -f` but drops lines w/o consequence\n"
         "   -l, --list                      Parse the VCF header and list the annotation fields\n"
-        "   -p, --annot-prefix              Prefix of INFO annotations to be created after splitting the CSQ string\n"
+        "   -p, --annot-prefix STR          Before doing anything else, prepend STR to all CSQ fields to avoid tag name conflicts\n"
         "   -s, --select TR:CSQ             Select transcripts to extract by type and/or consequence severity. (See also -S and -x.)\n"
         "                                     TR, transcript:   worst,primary(*),all        [all]\n"
         "                                     CSQ, consequence: any,missense,missense+,etc  [any]\n"
         "                                     (*) Primary transcripts have the field \"CANONICAL\" set to \"YES\"\n"
         "   -S, --severity -|FILE           Pass \"-\" to print the default severity scale or FILE to override\n"
         "                                     the default scale\n"
+        "   -u, --allow-undef-tags          Print \".\" for undefined tags\n"
         "   -x, --drop-sites                Drop sites with none of the consequences matching the severity specified by -s.\n"
         "                                      This switch is intended for use with VCF/BCF output (i.e. -f not given).\n"
         "Common options:\n"
@@ -210,11 +213,13 @@ static const char *usage_text(void)
         "   -i, --include EXPR              Include sites and samples for which the expression is true\n"
         "       --no-version                Do not append version and command line to the header\n"
         "   -o, --output FILE               Output file name [stdout]\n"
-        "   -O, --output-type b|u|z|v       b: compressed BCF, u: uncompressed BCF, z: compressed VCF or text, v: uncompressed VCF or text [v]\n"
+        "   -O, --output-type u|b|v|z[0-9]  u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n"
         "   -r, --regions REG               Restrict to comma-separated list of regions\n"
         "   -R, --regions-file FILE         Restrict to regions listed in a file\n"
+        "       --regions-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n"
         "   -t, --targets REG               Similar to -r but streams rather than index-jumps\n"
         "   -T, --targets-file FILE         Similar to -R but streams rather than index-jumps\n"
+        "       --targets-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n"
         "\n"
         "Examples:\n"
         "   # List available fields of the INFO/CSQ annotation\n"
@@ -244,6 +249,8 @@ static const char *usage_text(void)
         "   # the -x switch. See the online documentation referenced above for more examples.\n"
         "   bcftools +split-vep -c gnomAD_AF:Float -s :missense    file.vcf.gz\n"
         "   bcftools +split-vep -c gnomAD_AF:Float -s :missense -x file.vcf.gz\n"
+        "\n"
+        "   See also http://samtools.github.io/bcftools/howtos/plugin.split-vep.html\n"
         "\n";
 }
 
@@ -377,15 +384,30 @@ static int query_has_field(char *fmt, char *field, kstring_t *str)
     }
     return 1;
 }
+char *strdup_annot_prefix(args_t *args, const char *str)
+{
+    if ( !args->annot_prefix ) return strdup(str);
+    int str_len = strlen(str);
+    int prefix_len = strlen(args->annot_prefix);
+    char *out = calloc(str_len+prefix_len+1,1);
+    memcpy(out,args->annot_prefix,prefix_len);
+    memcpy(out+prefix_len,str,str_len);
+    return out;
+}
 static void init_data(args_t *args)
 {
     args->sr = bcf_sr_init();
     if ( args->regions )
     {
         args->sr->require_index = 1;
+        bcf_sr_set_opt(args->sr,BCF_SR_REGIONS_OVERLAP,args->regions_overlap);
         if ( bcf_sr_set_regions(args->sr, args->regions, args->regions_is_file)<0 ) error("Failed to read the regions: %s\n",args->regions);
     }
-    if ( args->targets && bcf_sr_set_targets(args->sr, args->targets, args->targets_is_file, 0)<0 ) error("Failed to read the targets: %s\n",args->targets);
+    if ( args->targets )
+    {
+        bcf_sr_set_opt(args->sr,BCF_SR_TARGETS_OVERLAP,args->targets_overlap);
+        if ( bcf_sr_set_targets(args->sr, args->targets, args->targets_is_file, 0)<0 ) error("Failed to read the targets: %s\n",args->targets);
+    }
     if ( !bcf_sr_add_reader(args->sr,args->fname) ) error("Error: %s\n", bcf_sr_strerror(args->sr->errnum));
     args->hdr = bcf_sr_get_header(args->sr,0);
     args->hdr_out = bcf_hdr_dup(args->hdr);
@@ -402,18 +424,18 @@ static void init_data(args_t *args)
     while ( *ep )
     {
         char *bp = ep;
-        while ( *ep && *ep!='|' ) ep++;
+        while ( *ep && *ep!='|' && *ep!='"' && *ep!='(' ) ep++; // don't include bracket in "pos(1-based)"
         char tmp = *ep;
         *ep = 0;
         args->nfield++;
         args->field = (char**)realloc(args->field,args->nfield*sizeof(*args->field));
-        args->field[args->nfield-1] = strdup(bp);
+        args->field[args->nfield-1] = strdup_annot_prefix(args,bp);
         if ( !tmp ) break;
-        ep++;
+        *ep = tmp;
+        while ( *ep && *ep!='|' ) ep++;     // skip bracket in "pos(1-based)"
+        if ( *ep ) ep++;
     }
     if ( !args->nfield ) error("Could not parse Description of INFO/%s: %s\n", args->vep_tag,hrec->vals[ret]);
-    int len = strlen(args->field[args->nfield-1]);
-    if ( args->field[args->nfield-1][len-1]=='"' ) args->field[args->nfield-1][len-1] = 0;    // remove the trailing doublequote character
     args->field2idx = khash_str2int_init();
     int i,j;
     for (i=0; i<args->nfield; i++)
@@ -494,8 +516,13 @@ static void init_data(args_t *args)
     cols_destroy(cols);
 
     // The 'CANONICAL' column to look up severity, its name is hardwired for now
-    if ( args->select_tr==SELECT_TR_PRIMARY && khash_str2int_get(args->field2idx,"CANONICAL",&args->primary_id)!=0 )
-        error("The primary transcript was requested but the field \"CANONICAL\" is not present in INFO/%s: %s\n",args->vep_tag,hrec->vals[ret]);
+    if ( args->select_tr==SELECT_TR_PRIMARY )
+    {
+        char *tmp = strdup_annot_prefix(args,"CANONICAL");
+        if ( khash_str2int_get(args->field2idx,tmp,&args->primary_id)!=0 )
+            error("The primary transcript was requested but the field \"CANONICAL\" is not present in INFO/%s: %s\n",args->vep_tag,hrec->vals[ret]);
+        free(tmp);
+    }
 
     // Create a text output as with `bcftools query -f`. For this we need to determine the fields to be extracted
     // from the formatting expression
@@ -527,8 +554,10 @@ static void init_data(args_t *args)
     }
 
     // The "Consequence" column to look up severity, its name is hardwired for now
-    if ( khash_str2int_get(args->field2idx,"Consequence",&args->csq_idx)!=0 )
+    char *tmp = strdup_annot_prefix(args,"Consequence");
+    if ( khash_str2int_get(args->field2idx,tmp,&args->csq_idx)!=0 )
         error("The field \"Consequence\" is not present in INFO/%s: %s\n", args->vep_tag,hrec->vals[ret]);
+    free(tmp);
 
     // Columns to extract: given as names, 0-based indexes or ranges of indexes
     if ( args->column_str )
@@ -561,7 +590,8 @@ static void init_data(args_t *args)
                 ep = str.s; 
                 continue;
             }
-            if ( khash_str2int_get(args->field2idx, bp, &idx_beg)==0 )
+            char *tmp = strdup_annot_prefix(args, bp);
+            if ( khash_str2int_get(args->field2idx, bp, &idx_beg)==0 || khash_str2int_get(args->field2idx, tmp, &idx_beg)==0 )
                 idx_end = idx_beg;
             else if ( (tp=strrchr(bp,':')) )
             {
@@ -569,7 +599,9 @@ static void init_data(args_t *args)
                 if ( khash_str2int_get(args->field2idx, bp, &idx_beg)!=0 )
                 {
                     *tp = ':';
-                    error("No such column: \"%s\"\n", bp);
+                    tp = strrchr(tmp,':');
+                    *tp = 0;
+                    if ( khash_str2int_get(args->field2idx, tmp, &idx_beg)!=0 ) error("No such column: \"%s\"\n", bp);
                 }
                 idx_end = idx_beg;
                 *tp = ':';
@@ -599,6 +631,7 @@ static void init_data(args_t *args)
                     }
                     else if ( !strcmp(bp,args->vep_tag) )
                     {
+                        free(tmp);
                         args->raw_vep_request = 1;
                         if ( !keep ) break;
                         ep++;
@@ -608,6 +641,7 @@ static void init_data(args_t *args)
                         error("No such column: \"%s\"\n", bp);
                 }
             }
+            free(tmp);
 
             i = args->nannot;
             args->nannot += idx_end - idx_beg + 1;
@@ -624,18 +658,13 @@ static void init_data(args_t *args)
             ep++;
         }
         args->annot = (annot_t*)calloc(args->nannot,sizeof(*args->annot));
-        int len = args->annot_prefix ? strlen(args->annot_prefix) : 0;
         for (i=0; i<args->nannot; i++)
         {
             annot_t *ann = &args->annot[i];
             ann->type = types[i];
             ann->idx = j = column[i];
             ann->field = strdup(args->field[j]);
-            int clen = strlen(args->field[j]);
-            ann->tag = (char*)malloc(clen+len+1);
-            if ( len ) memcpy(ann->tag,args->annot_prefix,len);
-            memcpy(ann->tag+len,ann->field,clen);
-            ann->tag[len+clen] = 0;
+            ann->tag = strdup(args->field[j]);
             args->kstr.l = 0;
             const char *type = "String";
             if ( ann->type==BCF_HT_REAL ) type = "Float";
@@ -670,6 +699,7 @@ static void init_data(args_t *args)
         if ( !args->column_str && !args->select ) error("Error: No %s field selected in the formatting expression and -s not given: a typo?\n",args->vep_tag);
         args->convert = convert_init(args->hdr_out, NULL, 0, args->format_str);
         if ( !args->convert ) error("Could not parse the expression: %s\n", args->format_str);
+        if ( args->allow_undef_tags ) convert_set_option(args->convert, allow_undef_tags, 1);
     }
     if ( args->filter_str )
     {
@@ -1007,6 +1037,9 @@ int run(int argc, char **argv)
     args->output_type  = FT_VCF;
     args->vep_tag = "CSQ";
     args->record_cmd_line = 1;
+    args->regions_overlap = 1;
+    args->targets_overlap = 0;
+    args->clevel = -1;
     static struct option loptions[] =
     {
         {"drop-sites",no_argument,0,'x'},
@@ -1026,13 +1059,17 @@ int run(int argc, char **argv)
         {"output-type",required_argument,NULL,'O'},
         {"regions",1,0,'r'},
         {"regions-file",1,0,'R'},
+        {"regions-overlap",required_argument,NULL,3},
         {"targets",1,0,'t'},
         {"targets-file",1,0,'T'},
+        {"targets-overlap",required_argument,NULL,4},
         {"no-version",no_argument,NULL,2},
+        {"allow-undef-tags",no_argument,0,'u'},
         {NULL,0,NULL,0}
     };
     int c;
-    while ((c = getopt_long(argc, argv, "o:O:i:e:r:R:t:T:lS:s:c:p:a:f:dA:x",loptions,NULL)) >= 0)
+    char *tmp;
+    while ((c = getopt_long(argc, argv, "o:O:i:e:r:R:t:T:lS:s:c:p:a:f:dA:xu",loptions,NULL)) >= 0)
     {
         switch (c) 
         {
@@ -1052,6 +1089,7 @@ int run(int argc, char **argv)
             case 'S': args->severity = optarg; break;
             case 's': args->select = optarg; break;
             case 'l': args->list_hdr = 1; break;
+            case 'u': args->allow_undef_tags = 1; break;
             case 'e':
                 if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
                 args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
@@ -1069,9 +1107,30 @@ int run(int argc, char **argv)
                           case 'u': args->output_type = FT_BCF; break;
                           case 'z': args->output_type = FT_VCF_GZ; break;
                           case 'v': args->output_type = FT_VCF; break;
-                          default: error("The output type \"%s\" not recognised\n", optarg);
+                          default:
+                          {
+                              args->clevel = strtol(optarg,&tmp,10);
+                              if ( *tmp || args->clevel<0 || args->clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                          }
+                      }
+                      if ( optarg[1] )
+                      {
+                          args->clevel = strtol(optarg+1,&tmp,10);
+                          if ( *tmp || args->clevel<0 || args->clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
                       }
                       break;
+            case  3 :
+                if ( !strcasecmp(optarg,"0") ) args->regions_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) args->regions_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) args->regions_overlap = 2;
+                else error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  4 :
+                if ( !strcasecmp(optarg,"0") ) args->targets_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) args->targets_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) args->targets_overlap = 2;
+                else error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case 'h':
             case '?':
             default: error("%s", usage_text()); break;
@@ -1107,7 +1166,9 @@ int run(int argc, char **argv)
             args->fh_bgzf = bgzf_open(args->output_fname, args->output_type&FT_GZ ? "wg" : "wu");
         else
         {
-            args->fh_vcf = hts_open(args->output_fname, hts_bcf_wmode2(args->output_type,args->output_fname));
+            char wmode[8];
+            set_wmode(wmode,args->output_type,args->output_fname,args->clevel);
+            args->fh_vcf = hts_open(args->output_fname ? args->output_fname : "-", wmode);
             if ( args->record_cmd_line ) bcf_hdr_append_version(args->hdr_out, args->argc, args->argv, "bcftools_split-vep");
             if ( bcf_hdr_write(args->fh_vcf, args->hdr_out)!=0 ) error("Failed to write the header to %s\n", args->output_fname);
         }
