@@ -1,6 +1,6 @@
 /*  vcffilter.c -- Apply fixed-threshold filters.
 
-    Copyright (C) 2013-2014 Genome Research Ltd.
+    Copyright (C) 2013-2021 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -25,8 +25,10 @@ THE SOFTWARE.  */
 #include <stdio.h>
 #include <unistd.h>
 #include <getopt.h>
+#include <assert.h>
 #include <ctype.h>
 #include <string.h>
+#include <strings.h>
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -60,7 +62,8 @@ typedef struct _args_t
     char *soft_filter;  // drop failed sites or annotate FILTER column?
     int annot_mode;     // add to existing FILTER annotation or replace? Otherwise reset FILTER to PASS or leave as it is?
     int flt_fail, flt_pass;     // BCF ids of fail and pass filters
-    int snp_gap, indel_gap, IndelGap_id, SnpGap_id;
+    int snp_gap, snp_gap_type, indel_gap, IndelGap_id, SnpGap_id;
+    char *snp_gap_str;
     int32_t ntmpi, *tmpi, ntmp_ac, *tmp_ac;
     rbuf_t rbuf;
     bcf1_t **rbuf_lines;
@@ -68,16 +71,18 @@ typedef struct _args_t
     bcf_srs_t *files;
     bcf_hdr_t *hdr;
     htsFile *out_fh;
-    int output_type, n_threads;
+    int output_type, n_threads, clevel;
 
     char **argv, *output_fname, *targets_list, *regions_list;
-    int argc;
+    int argc, record_cmd_line;
 }
 args_t;
 
 static void init_data(args_t *args)
 {
-    args->out_fh = hts_open(args->output_fname,hts_bcf_wmode(args->output_type));
+    char wmode[8];
+    set_wmode(wmode,args->output_type,args->output_fname,args->clevel);
+    args->out_fh = hts_open(args->output_fname ? args->output_fname : "-", wmode);
     if ( args->out_fh == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
     if ( args->n_threads ) hts_set_threads(args->out_fh, args->n_threads);
 
@@ -129,7 +134,8 @@ static void init_data(args_t *args)
                 if ( tmp.s ) kputs(" and ", &tmp);
                 kputs("\"IndelGap\"", &tmp);
             }
-            fprintf(stderr,"Warning: using %s filter name instead of \"%s\"\n", tmp.s,args->soft_filter);
+            if ( strncmp(tmp.s+1,args->soft_filter,tmp.l-2) )
+                fprintf(stderr,"Warning: using %s filter name instead of \"%s\"\n", tmp.s,args->soft_filter);
             free(tmp.s);
         }
 
@@ -137,7 +143,7 @@ static void init_data(args_t *args)
         args->rbuf_lines = (bcf1_t**) calloc(args->rbuf.m, sizeof(bcf1_t*));
         if ( args->snp_gap )
         {
-            bcf_hdr_printf(args->hdr, "##FILTER=<ID=SnpGap,Description=\"SNP within %d bp of an indel\">", args->snp_gap);
+            bcf_hdr_printf(args->hdr, "##FILTER=<ID=SnpGap,Description=\"SNP within %d bp of %s\">", args->snp_gap,args->snp_gap_str);
             args->SnpGap_id = bcf_hdr_id2int(args->hdr, BCF_DT_ID, "SnpGap");
             assert( args->SnpGap_id>=0 );
         }
@@ -149,7 +155,7 @@ static void init_data(args_t *args)
         }
     }
 
-    bcf_hdr_append_version(args->hdr, args->argc, args->argv, "bcftools_filter");
+    if (args->record_cmd_line) bcf_hdr_append_version(args->hdr, args->argc, args->argv, "bcftools_filter");
 
     if ( args->filter_str )
         args->filter = filter_init(args->hdr, args->filter_str);
@@ -187,7 +193,7 @@ static void flush_buffer(args_t *args, int n)
                 if ( args->snp_gap && rec->d.flt[j]==args->SnpGap_id ) { pass = 0; break; }
             }
         }
-        if ( pass ) bcf_write1(args->out_fh, args->hdr, rec);
+        if ( pass && bcf_write1(args->out_fh, args->hdr, rec)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
     }
 }
 
@@ -216,9 +222,9 @@ static void buffered_filters(args_t *args, bcf1_t *line)
      */
 
     // To avoid additional data structure, we abuse bcf1_t's var and var_type records.
-    const int SnpGap_set     = VCF_OTHER<<1;
-    const int IndelGap_set   = VCF_OTHER<<2;
-    const int IndelGap_flush = VCF_OTHER<<3;
+    const int SnpGap_set     = 1 << (8*sizeof(int)/2);
+    const int IndelGap_set   = 1 << (8*sizeof(int)/2-1);
+    const int IndelGap_flush = 1 << (8*sizeof(int)/2-2);
 
     int var_type = 0, i;
     if ( line )
@@ -244,15 +250,8 @@ static void buffered_filters(args_t *args, bcf1_t *line)
         // output REF=CAGAGAGAGA, ALT=CAGAGAGAGAGA where REF=C,ALT=CGA could be
         // used. This filter is therefore more strict and may remove some valid
         // SNPs.
-        int len = 1;
-        if ( var_type & VCF_INDEL )
-        {
-            for (i=1; i<line->n_allele; i++)
-                if ( len < 1-line->d.var[i].n ) len = 1-line->d.var[i].n;
-        }
-
         // Set the REF allele's length to max deletion length or to 1 if a SNP or an insertion.
-        line->d.var[0].n = len;
+        line->d.var[0].n = line->rlen;
     }
 
     int k_flush = 1;
@@ -277,7 +276,7 @@ static void buffered_filters(args_t *args, bcf1_t *line)
         if ( k_flush || !line )
         {
             // Select the best indel from the cluster of k_flush indels
-            int k = 0, max_ac = -1, imax_ac = -1;
+            int k = 0, max_ac = -1, imax_ac = -1, max_qual = -1, imax_qual = -1;
             for (i=-1; rbuf_next(&args->rbuf,&i) && k<k_flush; )
             {
                 k++;
@@ -286,9 +285,10 @@ static void buffered_filters(args_t *args, bcf1_t *line)
                 hts_expand(int, rec->n_allele, args->ntmpi, args->tmpi);
                 int ret = bcf_calc_ac(args->hdr, rec, args->tmpi, BCF_UN_ALL);
                 if ( imax_ac==-1 || (ret && max_ac < args->tmpi[1]) ) { max_ac = args->tmpi[1]; imax_ac = i; }
+                if ( imax_qual==-1 || max_qual < rec->qual ) { max_qual = rec->qual; imax_qual = i; }
             }
 
-            // Filter all but the best indel (with max AF or first if AF not available)
+            // Filter all but the best indel (with the best QUAL, bigger AC, or take the first if neither QUAL nor AC are available)
             k = 0;
             for (i=-1; rbuf_next(&args->rbuf,&i) && k<k_flush; )
             {
@@ -296,7 +296,14 @@ static void buffered_filters(args_t *args, bcf1_t *line)
                 bcf1_t *rec = args->rbuf_lines[i];
                 if ( !(rec->d.var_type & IndelGap_set) ) continue;
                 rec->d.var_type |= IndelGap_flush;
-                if ( i!=imax_ac ) bcf_add_filter(args->hdr, args->rbuf_lines[i], args->IndelGap_id);
+
+                int do_filter = 0;
+                if ( max_qual>0 )
+                {
+                    if ( i!=imax_qual ) do_filter = 1;
+                }
+                else if ( i!=imax_ac ) do_filter = 1;
+                if ( do_filter ) bcf_add_filter(args->hdr, args->rbuf_lines[i], args->IndelGap_id);
             }
         }
     }
@@ -319,13 +326,13 @@ static void buffered_filters(args_t *args, bcf1_t *line)
             int rec_to  = rec->pos + rec->d.var[0].n - 1;   // last position affected by the variant
             if ( rec_to + args->snp_gap < last_from )
                 j_flush++;
-            else if ( (var_type & VCF_INDEL) && (rec->d.var_type & VCF_SNP) && !(rec->d.var_type & SnpGap_set) )
+            else if ( (var_type & args->snp_gap_type) && (rec->d.var_type & VCF_SNP) && !(rec->d.var_type & SnpGap_set) )
             {
                 // this SNP has not been SnpGap-filtered yet
                 rec->d.var_type |= SnpGap_set;
                 bcf_add_filter(args->hdr, rec, args->SnpGap_id);
             }
-            else if ( (var_type & VCF_SNP) && (rec->d.var_type & VCF_INDEL) )
+            else if ( (var_type & VCF_SNP) && (rec->d.var_type & args->snp_gap_type) )
             {
                 // the line which we are adding is a SNP and needs to be filtered
                 line->d.var_type |= SnpGap_set;
@@ -403,20 +410,23 @@ static void usage(args_t *args)
     fprintf(stderr, "Usage:   bcftools filter [options] <in.vcf.gz>\n");
     fprintf(stderr, "\n");
     fprintf(stderr, "Options:\n");
-    fprintf(stderr, "    -e, --exclude <expr>          exclude sites for which the expression is true (see man page for details)\n");
-    fprintf(stderr, "    -g, --SnpGap <int>            filter SNPs within <int> base pairs of an indel\n");
-    fprintf(stderr, "    -G, --IndelGap <int>          filter clusters of indels separated by <int> or fewer base pairs allowing only one to pass\n");
-    fprintf(stderr, "    -i, --include <expr>          include only sites for which the expression is true (see man page for details\n");
-    fprintf(stderr, "    -m, --mode [+x]               \"+\": do not replace but add to existing FILTER; \"x\": reset filters at sites which pass\n");
-    fprintf(stderr, "    -o, --output <file>           write output to a file [standard output]\n");
-    fprintf(stderr, "    -O, --output-type <b|u|z|v>   b: compressed BCF, u: uncompressed BCF, z: compressed VCF, v: uncompressed VCF [v]\n");
-    fprintf(stderr, "    -r, --regions <region>        restrict to comma-separated list of regions\n");
-    fprintf(stderr, "    -R, --regions-file <file>     restrict to regions listed in a file\n");
-    fprintf(stderr, "    -s, --soft-filter <string>    annotate FILTER column with <string> or unique filter name (\"Filter%%d\") made up by the program (\"+\")\n");
-    fprintf(stderr, "    -S, --set-GTs <.|0>           set genotypes of failed samples to missing (.) or ref (0)\n");
-    fprintf(stderr, "    -t, --targets <region>        similar to -r but streams rather than index-jumps\n");
-    fprintf(stderr, "    -T, --targets-file <file>     similar to -R but streams rather than index-jumps\n");
-    fprintf(stderr, "        --threads <int>           number of extra output compression threads [0]\n");
+    fprintf(stderr, "    -e, --exclude EXPR             Exclude sites for which the expression is true (see man page for details)\n");
+    fprintf(stderr, "    -g, --SnpGap INT[:TYPE]        Filter SNPs within <int> base pairs of an indel (the default) or any combination of indel,mnp,bnd,other,overlap\n");
+    fprintf(stderr, "    -G, --IndelGap INT             Filter clusters of indels separated by <int> or fewer base pairs allowing only one to pass\n");
+    fprintf(stderr, "    -i, --include EXPR             Include only sites for which the expression is true (see man page for details\n");
+    fprintf(stderr, "    -m, --mode [+x]                \"+\": do not replace but add to existing FILTER; \"x\": reset filters at sites which pass\n");
+    fprintf(stderr, "        --no-version               Do not append version and command line to the header\n");
+    fprintf(stderr, "    -o, --output FILE              Write output to a file [standard output]\n");
+    fprintf(stderr, "    -O, --output-type u|b|v|z[0-9] u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n");
+    fprintf(stderr, "    -r, --regions REGION           Restrict to comma-separated list of regions\n");
+    fprintf(stderr, "    -R, --regions-file FILE        Restrict to regions listed in a file\n");
+    fprintf(stderr, "        --regions-overlap 0|1|2    Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(stderr, "    -s, --soft-filter STRING       Annotate FILTER column with <string> or unique filter name (\"Filter%%d\") made up by the program (\"+\")\n");
+    fprintf(stderr, "    -S, --set-GTs .|0              Set genotypes of failed samples to missing (.) or ref (0)\n");
+    fprintf(stderr, "    -t, --targets REGION           Similar to -r but streams rather than index-jumps\n");
+    fprintf(stderr, "    -T, --targets-file FILE        Similar to -R but streams rather than index-jumps\n");
+    fprintf(stderr, "        --targets-overlap 0|1|2    Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
+    fprintf(stderr, "        --threads INT              Use multithreading with <int> worker threads [0]\n");
     fprintf(stderr, "\n");
     exit(1);
 }
@@ -430,7 +440,11 @@ int main_vcffilter(int argc, char *argv[])
     args->output_fname = "-";
     args->output_type = FT_VCF;
     args->n_threads = 0;
+    args->record_cmd_line = 1;
+    args->clevel = -1;
     int regions_is_file = 0, targets_is_file = 0;
+    int regions_overlap = 1;
+    int targets_overlap = 0;
 
     static struct option loptions[] =
     {
@@ -441,21 +455,46 @@ int main_vcffilter(int argc, char *argv[])
         {"include",required_argument,NULL,'i'},
         {"targets",required_argument,NULL,'t'},
         {"targets-file",required_argument,NULL,'T'},
+        {"targets-overlap",required_argument,NULL,4},
         {"regions",required_argument,NULL,'r'},
         {"regions-file",required_argument,NULL,'R'},
+        {"regions-overlap",required_argument,NULL,3},
         {"output",required_argument,NULL,'o'},
         {"output-type",required_argument,NULL,'O'},
         {"threads",required_argument,NULL,9},
         {"SnpGap",required_argument,NULL,'g'},
         {"IndelGap",required_argument,NULL,'G'},
+        {"no-version",no_argument,NULL,8},
         {NULL,0,NULL,0}
     };
     char *tmp;
     while ((c = getopt_long(argc, argv, "e:i:t:T:r:R:h?s:m:o:O:g:G:S:",loptions,NULL)) >= 0) {
         switch (c) {
-            case 'g': 
+            case 'g':
                 args->snp_gap = strtol(optarg,&tmp,10); 
-                if ( *tmp ) error("Could not parse argument: --SnpGap %s\n", optarg);
+                if ( *tmp && *tmp!=':' ) error("Could not parse argument: --SnpGap %s\n", optarg);
+                if ( *tmp==':' )
+                {
+                    args->snp_gap_str = tmp+1;
+                    int i,n;
+                    char **keys = hts_readlist(tmp+1,0,&n);
+                    for(i=0; i<n; i++)
+                    {
+                        if ( !strcasecmp(keys[i],"indel") ) args->snp_gap_type |= VCF_INDEL;
+                        else if ( !strcasecmp(keys[i],"mnp") ) args->snp_gap_type |= VCF_MNP;
+                        else if ( !strcasecmp(keys[i],"bnd") ) args->snp_gap_type |= VCF_BND;
+                        else if ( !strcasecmp(keys[i],"other") ) args->snp_gap_type |= VCF_OTHER;
+                        else if ( !strcasecmp(keys[i],"overlap") ) args->snp_gap_type |= VCF_OVERLAP;
+                        else error("Could not parse \"%s\" in \"--SnpGap %s\"\n", keys[i], optarg);
+                        free(keys[i]);
+                    }
+                    if ( n ) free(keys);
+                }
+                else
+                {
+                    args->snp_gap_type = VCF_INDEL;
+                    args->snp_gap_str = "indel";
+                }
                 break;
             case 'G':
                 args->indel_gap = strtol(optarg,&tmp,10);
@@ -468,7 +507,16 @@ int main_vcffilter(int argc, char *argv[])
                     case 'u': args->output_type = FT_BCF; break;
                     case 'z': args->output_type = FT_VCF_GZ; break;
                     case 'v': args->output_type = FT_VCF; break;
-                    default: error("The output type \"%s\" not recognised\n", optarg);
+                    default:
+                    {
+                        args->clevel = strtol(optarg,&tmp,10);
+                        if ( *tmp || args->clevel<0 || args->clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                    }
+                }
+                if ( optarg[1] )
+                {
+                    args->clevel = strtol(optarg+1,&tmp,10);
+                    if ( *tmp || args->clevel<0 || args->clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
                 }
                 break;
             case 's': args->soft_filter = optarg; break;
@@ -480,16 +528,33 @@ int main_vcffilter(int argc, char *argv[])
             case 'T': args->targets_list = optarg; targets_is_file = 1; break;
             case 'r': args->regions_list = optarg; break;
             case 'R': args->regions_list = optarg; regions_is_file = 1; break;
-            case 'e': args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
-            case 'i': args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
+            case 'e':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
+            case 'i':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'S':
                 if ( !strcmp(".",optarg) ) args->set_gts = SET_GTS_MISSING;
                 else if ( !strcmp("0",optarg) ) args->set_gts = SET_GTS_REF;
                 else error("The argument to -S not recognised: %s\n", optarg);
                 break;
             case  9 : args->n_threads = strtol(optarg, 0, 0); break;
+            case  8 : args->record_cmd_line = 0; break;
+            case  3 :
+                if ( !strcasecmp(optarg,"0") ) regions_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) regions_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) regions_overlap = 2;
+                else error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  4 :
+                if ( !strcasecmp(optarg,"0") ) targets_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) targets_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) targets_overlap = 2;
+                else error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case 'h':
-            case '?': usage(args);
+            case '?': usage(args); break;
             default: error("Unknown argument: %s\n", optarg);
         }
     }
@@ -507,6 +572,7 @@ int main_vcffilter(int argc, char *argv[])
     if ( args->regions_list )
     {
         args->files->require_index = 1;
+        bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
         if ( bcf_sr_set_regions(args->files, args->regions_list, regions_is_file)<0 )
             error("Failed to read the regions: %s\n", args->regions_list);
     }
@@ -517,19 +583,21 @@ int main_vcffilter(int argc, char *argv[])
         kputs(argv[optind+1],&tmp);
         for (i=optind+2; i<argc; i++) { kputc(',',&tmp); kputs(argv[i],&tmp); }
         args->files->require_index = 1;
+        bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
         if ( bcf_sr_set_regions(args->files, tmp.s, regions_is_file)<0 )
             error("Failed to read the regions: %s\n", args->regions_list);
         free(tmp.s);
     }
     if ( args->targets_list )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_TARGETS_OVERLAP,targets_overlap);
         if ( bcf_sr_set_targets(args->files, args->targets_list,targets_is_file, 0)<0 )
             error("Failed to read the targets: %s\n", args->targets_list);
     }
-    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to open %s: %s\n", fname,bcf_sr_strerror(args->files->errnum));
+    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to read from %s: %s\n", !strcmp("-",fname)?"standard input":fname,bcf_sr_strerror(args->files->errnum));
 
     init_data(args);
-    bcf_hdr_write(args->out_fh, args->hdr);
+    if ( bcf_hdr_write(args->out_fh, args->hdr)!=0 ) error("[%s] Error: cannot write the header to %s\n", __func__,args->output_fname);
     while ( bcf_sr_next_line(args->files) )
     {
         bcf1_t *line = bcf_sr_get_line(args->files, 0);
@@ -553,14 +621,16 @@ int main_vcffilter(int argc, char *argv[])
             }
             if ( args->set_gts ) set_genotypes(args, line, pass);
             if ( !args->rbuf_lines )
-                bcf_write1(args->out_fh, args->hdr, line);
+            {
+                if ( bcf_write1(args->out_fh, args->hdr, line)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
+            }
             else
                 buffered_filters(args, line);
         }
     }
     buffered_filters(args, NULL);
 
-    hts_close(args->out_fh);
+    if ( hts_close(args->out_fh)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
     destroy_data(args);
     bcf_sr_destroy(args->files);
     free(args);
