@@ -1,6 +1,6 @@
 /*  vcfplugin.c -- plugin modules for operating on VCF/BCF files.
 
-    Copyright (C) 2013-2016 Genome Research Ltd.
+    Copyright (C) 2013-2021 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -22,6 +22,7 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.  */
 
+#include "config.h"
 #include <stdio.h>
 #include <strings.h>
 #include <unistd.h>
@@ -37,11 +38,17 @@ THE SOFTWARE.  */
 #include <htslib/synced_bcf_reader.h>
 #include <htslib/kseq.h>
 #include <htslib/khash_str2int.h>
+#ifdef _WIN32
+#include <windows.h>
+#else
 #include <dlfcn.h>
+#endif
 #include "bcftools.h"
 #include "vcmp.h"
 #include "filter.h"
 
+#ifdef ENABLE_BCF_PLUGINS
+
 typedef struct _plugin_t plugin_t;
 
 /**
@@ -130,7 +137,7 @@ typedef struct _args_t
     bcf_srs_t *files;
     bcf_hdr_t *hdr, *hdr_out;
     htsFile *out_fh;
-    int output_type, n_threads;
+    int output_type, n_threads, clevel;
 
     filter_t *filter;
     char *filter_str;
@@ -141,7 +148,7 @@ typedef struct _args_t
     char **plugin_paths;
 
     char **argv, *output_fname, *regions_list, *targets_list;
-    int argc, drop_header, verbose, record_cmd_line;
+    int argc, drop_header, verbose, record_cmd_line, plist_only;
 }
 args_t;
 
@@ -151,7 +158,7 @@ static void add_plugin_paths(args_t *args, const char *path)
 {
     while (1)
     {
-        size_t len = strcspn(path, ":");
+        size_t len = strcspn(path, HTS_PATH_SEPARATOR_STR);
 
         if ( len == 0 )
         {
@@ -171,7 +178,7 @@ static void add_plugin_paths(args_t *args, const char *path)
                 args->plugin_paths = (char**) realloc(args->plugin_paths,sizeof(char*)*(args->nplugin_paths+1));
                 args->plugin_paths[args->nplugin_paths] = dir;
                 args->nplugin_paths++;
-                if ( args->verbose > 1 ) fprintf(stderr, "plugin directory %s .. ok\n", dir);
+                if ( args->verbose > 1 && strcmp(".",dir) ) fprintf(stderr, "plugin directory %s .. ok\n", dir);
             }
             else
             {
@@ -182,7 +189,7 @@ static void add_plugin_paths(args_t *args, const char *path)
         }
 
         path += len;
-        if ( *path == ':' ) path++;
+        if ( *path == HTS_PATH_SEPARATOR_CHAR ) path++;
         else break;
     }
 }
@@ -204,46 +211,86 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 
     void *handle;
     char *tmp;
-    if ( fname[0]!='/' )    // not an absolute path
+    int is_absolute_path = 0;
+#ifdef _WIN32
+    // Windows accepts both forward slash (/) and backslash (\) as folder separator
+    // and can have any path prefixed by the drive letter and a colon (:).
+    if ( fname[0]=='/' || fname[0]=='\\') is_absolute_path = 1;
+    else if ( fname[0] && fname[1]==':' && (fname[2]=='/' || fname[2]=='\\') ) is_absolute_path = 1;
+#else
+    if ( fname[0]=='/' ) is_absolute_path = 1;
+#endif
+
+    kstring_t err = {0,0,0};
+    if ( !is_absolute_path )
     {
         int i;
         for (i=0; i<args->nplugin_paths; i++)
         {
-            tmp = msprintf("%s/%s.so", args->plugin_paths[i],fname);
+            tmp = msprintf("%s/%s%s", args->plugin_paths[i], fname, PLUGIN_EXT);
+#ifdef _WIN32
+            handle = LoadLibraryA(tmp);
+#else
             handle = dlopen(tmp, RTLD_NOW); // valgrind complains about unfreed memory, not our problem though
-            if ( args->verbose > 1 )
-            {
-                if ( !handle ) fprintf(stderr,"%s:\n\tdlopen   .. %s\n", tmp,dlerror());
-                else fprintf(stderr,"%s:\n\tdlopen   .. ok\n", tmp);
-            }
+#endif
+            if ( !handle )
+#ifdef _WIN32
+                ksprintf(&err,"LoadLibraryA   .. %lu\n", GetLastError());
+#else
+                ksprintf(&err,"%s:\n\tdlopen   .. %s\n", tmp,dlerror());
+#endif
+            else if ( args->verbose > 1 )
+                fprintf(stderr,"%s:\n\tplugin open   .. ok\n", tmp);
             free(tmp);
             if ( handle ) return handle;
         }
     }
 
+#ifdef _WIN32
+    handle = LoadLibraryA(fname);
+#else
     handle = dlopen(fname, RTLD_NOW);
-    if ( args->verbose > 1 )
-    {
-        if ( !handle ) fprintf(stderr,"%s:\n\tdlopen   .. %s\n", fname,dlerror());
-        else fprintf(stderr,"%s:\n\tdlopen   .. ok\n", fname);
-    }
+#endif
+    if ( !handle )
+#ifdef _WIN32
+        ksprintf(&err,"LoadLibraryA   .. %lu\n", GetLastError());
+#else
+        ksprintf(&err,"%s:\n\tdlopen   .. %s\n", fname,dlerror());
+#endif
+    else if ( args->verbose > 1 )
+        fprintf(stderr,"%s:\n\tplugin open   .. ok\n", fname);
+
+    if ( !handle && (!args->plist_only || args->verbose>1) )
+        fprintf(stderr,"%s",err.s);
+    free(err.s);
 
     return handle;
 }
 
-static void print_plugin_usage_hint(void)
+static void print_plugin_usage_hint(const char *name)
 {
-    fprintf(stderr, "\nNo functional bcftools plugins were found");
-    if ( !getenv("BCFTOOLS_PLUGINS") )
-        fprintf(stderr,". The environment variable BCFTOOLS_PLUGINS is not set.\n\n");
+    if ( name )
+        fprintf(stderr, "\nThe bcftools plugin \"%s\" was not found or is not functional", name);
     else
+        fprintf(stderr, "\nNo functional bcftools plugins were found");
+    if ( !getenv("BCFTOOLS_PLUGINS") )
+    {
+        fprintf(stderr,". The environment variable BCFTOOLS_PLUGINS is not set");
+#ifdef PLUGINPATH
+        fprintf(stderr,"\nand no usable plugins were found in %s", PLUGINPATH);
+#endif
+        fprintf(stderr,".\n\n");
+    }
+    else
+    {
         fprintf(stderr,
                 " in\n\tBCFTOOLS_PLUGINS=\"%s\".\n\n"
                 "- Is the plugin path correct?\n\n"
-                "- Run \"bcftools plugin -lv\" for more detailed error output.\n"
+                "- Run \"bcftools plugin -l\" or \"bcftools plugin -lvv\" for a list of available plugins.\n"
                 "\n",
                 getenv("BCFTOOLS_PLUGINS")
                );
+    }
 }
 
 static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugin_t *plugin)
@@ -255,12 +302,61 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
     {
         if ( exit_on_error )
         {
-            print_plugin_usage_hint();
+            print_plugin_usage_hint(fname);
             error("Could not load \"%s\".\n\n", fname);
         }
         return -1;
     }
 
+#ifdef _WIN32
+    plugin->init = (dl_init_f) GetProcAddress(plugin->handle, "init");
+    if ( plugin->init && args->verbose > 1 ) fprintf(stderr,"\tinit     .. ok\n");
+
+    plugin->run = (dl_run_f) GetProcAddress(plugin->handle, "run");
+    if ( plugin->run && args->verbose > 1 ) fprintf(stderr,"\trun     .. ok\n");
+
+    if ( !plugin->init && !plugin->run )
+    {
+        if ( exit_on_error ) error("Could not initialize %s, neither run or init found \n", plugin->name);
+        else if ( args->verbose > 1 ) fprintf(stderr,"\tinit/run .. not found\n");
+        return -1;
+    }
+
+    plugin->version = (dl_version_f) GetProcAddress(plugin->handle, "version");
+    if ( !plugin->version )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: version string not found\n", plugin->name);
+        else if ( args->verbose > 1 ) fprintf(stderr,"\tversion  .. not found\n");
+        return -1;
+    }
+
+    plugin->about = (dl_about_f) GetProcAddress(plugin->handle, "about");
+    if ( !plugin->about )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: about string not found\n", plugin->name);
+        return -1;
+    }
+
+    plugin->usage = (dl_about_f) GetProcAddress(plugin->handle, "usage");
+    if ( !plugin->usage )
+        plugin->usage = plugin->about;
+
+    if ( plugin->run ) return 0;
+
+    plugin->process = (dl_process_f) GetProcAddress(plugin->handle, "process");
+    if ( !plugin->process )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: process method not found\n", plugin->name);
+        return -1;
+    }
+
+    plugin->destroy = (dl_destroy_f) GetProcAddress(plugin->handle, "destroy");
+    if ( !plugin->destroy )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: destroy method not found\n", plugin->name);
+        return -1;
+    }
+#else
     dlerror();
     plugin->init = (dl_init_f) dlsym(plugin->handle, "init");
     char *ret = dlerror();
@@ -322,16 +418,14 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
         if ( exit_on_error ) error("Could not initialize %s: %s\n", plugin->name, ret);
         return -1;
     }
+#endif
 
     return 0;
 }
 
-static void init_plugin(args_t *args)
+static void check_version(args_t *args)
 {
     static int warned_bcftools = 0, warned_htslib = 0;
-
-    int ret = args->plugin.init(args->plugin.argc,args->plugin.argv,args->hdr,args->hdr_out);
-    if ( ret<0 ) error("The plugin exited with an error.\n");
     const char *bver, *hver;
     args->plugin.version(&bver, &hver);
     if ( strcmp(bver,bcftools_version()) && !warned_bcftools )
@@ -344,6 +438,13 @@ static void init_plugin(args_t *args)
         fprintf(stderr,"WARNING: htslib version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", hts_version(),args->plugin.name,hver);
         warned_htslib = 1;
     }
+}
+
+static void init_plugin(args_t *args)
+{
+    int ret = args->plugin.init(args->plugin.argc,args->plugin.argv,args->hdr,args->hdr_out);
+    if ( ret<0 ) error("The plugin exited with an error.\n");
+    check_version(args);
     args->drop_header += ret;
 }
 
@@ -362,6 +463,7 @@ static int list_plugins(args_t *args)
     init_plugin_paths(args);
 
     kstring_t str = {0,0,0};
+    int plugin_ext_len = strlen(PLUGIN_EXT);
     int i;
     for (i=0; i<args->nplugin_paths; i++)
     {
@@ -372,7 +474,7 @@ static int list_plugins(args_t *args)
         while ( (ep=readdir(dp)) )
         {
             int len = strlen(ep->d_name);
-            if ( strcasecmp(".so",ep->d_name+len-3) ) continue;
+            if ( strcasecmp(PLUGIN_EXT,ep->d_name+len-plugin_ext_len) ) continue;
             str.l = 0;
             ksprintf(&str,"%s/%s", args->plugin_paths[i],ep->d_name);
             hts_expand(plugin_t, nplugins+1, mplugins, plugins);
@@ -402,7 +504,7 @@ static int list_plugins(args_t *args)
         if ( args->verbose ) printf("\n");
     }
     else
-        print_plugin_usage_hint();
+        print_plugin_usage_hint(NULL);
     free(str.s);
     return nplugins ? 0 : 1;
 }
@@ -420,10 +522,12 @@ static void init_data(args_t *args)
     if (args->record_cmd_line) bcf_hdr_append_version(args->hdr_out, args->argc, args->argv, "bcftools_plugin");
     if ( !args->drop_header )
     {
-        args->out_fh = hts_open(args->output_fname,hts_bcf_wmode(args->output_type));
+        char wmode[8];
+        set_wmode(wmode,args->output_type,args->output_fname,args->clevel);
+        args->out_fh = hts_open(args->output_fname ? args->output_fname : "-", wmode);
         if ( args->out_fh == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
         if ( args->n_threads ) hts_set_threads(args->out_fh, args->n_threads);
-        bcf_hdr_write(args->out_fh, args->hdr_out);
+        if ( bcf_hdr_write(args->out_fh, args->hdr_out)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
     }
 }
 
@@ -431,7 +535,11 @@ static void destroy_data(args_t *args)
 {
     free(args->plugin.name);
     if ( args->plugin.destroy ) args->plugin.destroy();
+#ifdef _WIN32
+    FreeLibrary(args->plugin.handle);
+#else
     dlclose(args->plugin.handle);
+#endif
     if ( args->hdr_out ) bcf_hdr_destroy(args->hdr_out);
     if ( args->nplugin_paths>0 )
     {
@@ -441,7 +549,7 @@ static void destroy_data(args_t *args)
     }
     if ( args->filter )
         filter_destroy(args->filter);
-    if (args->out_fh) hts_close(args->out_fh);
+    if (args->out_fh && hts_close(args->out_fh)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
 }
 
 static void usage(args_t *args)
@@ -452,22 +560,24 @@ static void usage(args_t *args)
     fprintf(stderr, "         bcftools +name [OPTIONS] <file>  [-- PLUGIN_OPTIONS]\n");
     fprintf(stderr, "\n");
     fprintf(stderr, "VCF input options:\n");
-    fprintf(stderr, "   -e, --exclude <expr>        exclude sites for which the expression is true\n");
-    fprintf(stderr, "   -i, --include <expr>        select sites for which the expression is true\n");
-    fprintf(stderr, "   -r, --regions <region>      restrict to comma-separated list of regions\n");
-    fprintf(stderr, "   -R, --regions-file <file>   restrict to regions listed in a file\n");
-    fprintf(stderr, "   -t, --targets <region>      similar to -r but streams rather than index-jumps\n");
-    fprintf(stderr, "   -T, --targets-file <file>   similar to -R but streams rather than index-jumps\n");
+    fprintf(stderr, "   -e, --exclude EXPR             Exclude sites for which the expression is true\n");
+    fprintf(stderr, "   -i, --include EXPR             Select sites for which the expression is true\n");
+    fprintf(stderr, "   -r, --regions REGION           Restrict to comma-separated list of regions\n");
+    fprintf(stderr, "   -R, --regions-file FILE        Restrict to regions listed in a file\n");
+    fprintf(stderr, "        --regions-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(stderr, "   -t, --targets REGION           Similar to -r but streams rather than index-jumps\n");
+    fprintf(stderr, "   -T, --targets-file FILE        Similar to -R but streams rather than index-jumps\n");
+    fprintf(stderr, "        --targets-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
     fprintf(stderr, "VCF output options:\n");
-    fprintf(stderr, "       --no-version            do not append version and command line to the header\n");
-    fprintf(stderr, "   -o, --output <file>         write output to a file [standard output]\n");
-    fprintf(stderr, "   -O, --output-type <type>    'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n");
-    fprintf(stderr, "       --threads <int>         number of extra output compression threads [0]\n");
+    fprintf(stderr, "       --no-version               Do not append version and command line to the header\n");
+    fprintf(stderr, "   -o, --output FILE              Write output to a file [standard output]\n");
+    fprintf(stderr, "   -O, --output-type u|b|v|z[0-9] u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n");
+    fprintf(stderr, "       --threads INTT             Use multithreading with <int> worker threads [0]\n");
     fprintf(stderr, "Plugin options:\n");
-    fprintf(stderr, "   -h, --help                  list plugin's options\n");
-    fprintf(stderr, "   -l, --list-plugins          list available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
-    fprintf(stderr, "   -v, --verbose               print verbose information, -vv increases verbosity\n");
-    fprintf(stderr, "   -V, --version               print version string and exit\n");
+    fprintf(stderr, "   -h, --help                     List plugin's options\n");
+    fprintf(stderr, "   -l, --list-plugins             List available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
+    fprintf(stderr, "   -v, --verbose                  Print verbose information, -vv increases verbosity\n");
+    fprintf(stderr, "   -V, --version                  Print version string and exit\n");
     fprintf(stderr, "\n");
     exit(1);
 }
@@ -503,10 +613,12 @@ int main_plugin(int argc, char *argv[])
     args->n_threads = 0;
     args->record_cmd_line = 1;
     args->nplugin_paths = -1;
-    int regions_is_file = 0, targets_is_file = 0, plist_only = 0, usage_only = 0, version_only = 0;
+    args->clevel = -1;
+    int regions_is_file = 0, targets_is_file = 0, usage_only = 0, version_only = 0;
+    int regions_overlap = 1;
+    int targets_overlap = 0;
 
     if ( argc==1 ) usage(args);
-
     char *plugin_name = NULL;
     if ( argv[1][0]!='-' )
     {
@@ -517,6 +629,7 @@ int main_plugin(int argc, char *argv[])
         load_plugin(args, plugin_name, 1, &args->plugin);
         if ( args->plugin.run )
         {
+            check_version(args);
             int ret = args->plugin.run(argc, argv);
             destroy_data(args);
             free(args);
@@ -537,11 +650,14 @@ int main_plugin(int argc, char *argv[])
         {"exclude",required_argument,NULL,'e'},
         {"regions",required_argument,NULL,'r'},
         {"regions-file",required_argument,NULL,'R'},
+        {"regions-overlap",required_argument,NULL,1},
         {"targets",required_argument,NULL,'t'},
         {"targets-file",required_argument,NULL,'T'},
+        {"targets-overlap",required_argument,NULL,2},
         {"no-version",no_argument,NULL,8},
         {NULL,0,NULL,0}
     };
+    char *tmp;
     while ((c = getopt_long(argc, argv, "h?o:O:r:R:t:T:li:e:vV",loptions,NULL)) >= 0)
     {
         switch (c) {
@@ -554,16 +670,41 @@ int main_plugin(int argc, char *argv[])
                     case 'u': args->output_type = FT_BCF; break;
                     case 'z': args->output_type = FT_VCF_GZ; break;
                     case 'v': args->output_type = FT_VCF; break;
-                    default: error("The output type \"%s\" not recognised\n", optarg);
+                    default:
+                    {
+                        args->clevel = strtol(optarg,&tmp,10);
+                        if ( *tmp || args->clevel<0 || args->clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                    }
                 };
+                if ( optarg[1] )
+                {
+                    args->clevel = strtol(optarg+1,&tmp,10);
+                    if ( *tmp || args->clevel<0 || args->clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
+                }
                 break;
-            case 'e': args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
-            case 'i': args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
+            case 'e':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
+            case 'i':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'r': args->regions_list = optarg; break;
             case 'R': args->regions_list = optarg; regions_is_file = 1; break;
             case 't': args->targets_list = optarg; break;
             case 'T': args->targets_list = optarg; targets_is_file = 1; break;
-            case 'l': plist_only = 1; break;
+            case 'l': args->plist_only = 1; break;
+            case  1 :
+                if ( !strcasecmp(optarg,"0") ) regions_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) regions_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) regions_overlap = 2;
+                else error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  2 :
+                if ( !strcasecmp(optarg,"0") ) targets_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) targets_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) targets_overlap = 2;
+                else error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case  9 : args->n_threads = strtol(optarg, 0, 0); break;
             case  8 : args->record_cmd_line = 0; break;
             case '?':
@@ -571,8 +712,8 @@ int main_plugin(int argc, char *argv[])
             default: error("Unknown argument: %s\n", optarg);
         }
     }
-    if ( plist_only )  return list_plugins(args);
-    if ( usage_only && ! plugin_name ) usage(args);
+    if ( args->plist_only )  return list_plugins(args);
+    if ( !plugin_name ) usage(args);
 
     if ( version_only )
     {
@@ -593,12 +734,18 @@ int main_plugin(int argc, char *argv[])
     }
 
     char *fname = NULL;
-    if ( optind>=argc || argv[optind][0]=='-' )
+    if ( optind>=argc || (argv[optind][0]=='-' && argv[optind][1]) )
     {
-        if ( !isatty(fileno((FILE *)stdin)) ) fname = "-";  // reading from stdin
-        else usage(args);
         args->plugin.argc = argc - optind + 1;
         args->plugin.argv = argv + optind - 1;
+
+        if ( !isatty(fileno((FILE *)stdin)) ) fname = "-";  // reading from stdin
+        else if ( optind>=argc ) usage(args);
+        else
+        {
+            optind = 1;
+            init_plugin(args);
+        }
     }
     else
     {
@@ -611,16 +758,18 @@ int main_plugin(int argc, char *argv[])
     args->files = bcf_sr_init();
     if ( args->regions_list )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
         if ( bcf_sr_set_regions(args->files, args->regions_list, regions_is_file)<0 )
             error("Failed to read the regions: %s\n", args->regions_list);
     }
     if ( args->targets_list )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_TARGETS_OVERLAP,targets_overlap);
         if ( bcf_sr_set_targets(args->files, args->targets_list, targets_is_file, 0)<0 )
             error("Failed to read the targets: %s\n", args->targets_list);
         args->files->collapse |= COLLAPSE_SOME;
     }
-    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to open %s: %s\n", fname,bcf_sr_strerror(args->files->errnum));
+    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to read from %s: %s\n", !strcmp("-",fname)?"standard input":fname,bcf_sr_strerror(args->files->errnum));
 
     init_data(args);
     while ( bcf_sr_next_line(args->files) )
@@ -633,7 +782,11 @@ int main_plugin(int argc, char *argv[])
             if ( !pass ) continue;
         }
         line = args->plugin.process(line);
-        if ( line ) bcf_write1(args->out_fh, args->hdr_out, line);
+        if ( line )
+        {
+            if ( line->errcode ) error("[E::main_plugin] Unchecked error (%d), exiting\n",line->errcode);
+            if ( bcf_write1(args->out_fh, args->hdr_out, line)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
+        }
     }
     destroy_data(args);
     bcf_sr_destroy(args->files);
@@ -641,3 +794,13 @@ int main_plugin(int argc, char *argv[])
     return 0;
 }
 
+#else /* ENABLE_BCF_PLUGINS */
+
+int main_plugin(int argc, char *argv[])
+{
+    fprintf(stderr, "bcftools plugins are disabled.  To use them, you will need to rebuild\n"
+	    "bcftools from the source distribution with plugins enabled.\n");
+    return 1;
+}
+
+#endif /* ENABLE_BCF_PLUGINS */
